@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -28,24 +28,29 @@ from __future__ import unicode_literals, absolute_import
 
 import os
 import sys
+import json
 import time
 import platform
 import argparse
 import datetime
 import socket
 import shutil
+import subprocess
 import warnings
 import logging
 from getpass import getpass
 
 import six
+import humanize
 
 from rattail import __version__
 from rattail.core import Object
 from rattail.util import load_entry_points, load_object
-from rattail.console import Progress
-from rattail.config import make_config
+from rattail.progress import ConsoleProgress, SocketProgress
+from rattail.config import make_config, parse_list
 from rattail.util import progress_loop
+from rattail.time import make_utc
+from rattail.db.config import configure_versioning
 
 
 log = logging.getLogger(__name__)
@@ -82,6 +87,37 @@ def date_argument(string):
     return date
 
 
+def dict_argument(string):
+    """
+    Coerce the given string to a Python dictionary.  The string is assumed to
+    be JSON-encoded; this function merely invokes ``json.loads()`` on it.
+
+    This function is designed be used as the ``type`` parameter when calling
+    ``ArgumentParser.add_argument()``, e.g.::
+
+       parser = ArgumentParser()
+       parser.add_argument('--date', type=dict_argument)
+    """
+    try:
+        return json.loads(string)
+    except json.decoder.JSONDecodeError:
+        raise argparse.ArgumentTypeError("Argument must be valid JSON-encoded string")
+
+
+def list_argument(string):
+    """
+    Coerce the given string to a list of strings, splitting on whitespace
+    and/or commas.
+
+    This function is designed be used as the ``type`` parameter when calling
+    ``ArgumentParser.add_argument()``, e.g.::
+
+       parser = ArgumentParser()
+       parser.add_argument('--things', type=list_argument)
+    """
+    return parse_list(string)
+
+
 @six.python_2_unicode_compatible
 class Command(Object):
     """
@@ -143,8 +179,7 @@ See the file COPYING.txt for more information.
         Reference to the Python module which is to be used as the primary data
         model.  Defaults to ``rattail.db.model``.
         """
-        from rattail.db import model
-        return model
+        return self.config.get_model()
 
     def iter_subcommands(self):
         """
@@ -198,7 +233,10 @@ Commands:\n""".format(self))
         parser.add_argument('-n', '--no-init', action='store_true', default=False)
         parser.add_argument('--no-extend-config', dest='extend_config', action='store_false')
 
+        parser.add_argument('--verbose', action='store_true')
         parser.add_argument('-P', '--progress', action='store_true', default=False)
+        parser.add_argument('--progress-socket',
+                            help="Optional socket (e.g. localhost:8487) to which progress info should be written.")
         parser.add_argument('--runas', '-R', metavar='USERNAME',
                             help="Optional username to impersonate when running the command.  "
                             "This is only relevant for / used by certain commands.")
@@ -206,6 +244,17 @@ Commands:\n""".format(self))
                             help="Optional path to which STDOUT should be effectively redirected.")
         parser.add_argument('--stderr', metavar='PATH', type=argparse.FileType('w'),
                             help="Optional path to which STDERR should be effectively redirected.")
+
+        # data versioning
+        parser.add_argument('--versioning', action='store_true',
+                            help="Force *enable* of data versioning.  If set, then --no-versioning "
+                            "cannot also be set.  If neither is set, config will determine whether "
+                            "or not data versioning should be enabled.")
+        parser.add_argument('--no-versioning', action='store_true',
+                            help="Force *disable* of data versioning.  If set, then --versioning "
+                            "cannot also be set.  If neither is set, config will determine whether "
+                            "or not data versioning should be enabled.")
+
         parser.add_argument('-V', '--version', action='version',
                             version="%(prog)s {0}".format(self.version))
         parser.add_argument('command', nargs='*')
@@ -216,6 +265,11 @@ Commands:\n""".format(self))
             self.print_help()
             return
 
+        # TODO: can we make better args so this is handled by argparse somehow?
+        if args.versioning and args.no_versioning:
+            sys.stderr.write("Cannot pass both --versioning and --no-versioning\n")
+            sys.exit(1)
+
         # Show (sub)command help if so instructed, or unknown subcommand.
         cmd = args.command.pop(0)
         if cmd == 'help':
@@ -229,7 +283,12 @@ Commands:\n""".format(self))
             cmd = self.subcommands[cmd](parent=self)
             cmd.parser.print_help()
             return
-        elif cmd not in self.subcommands:
+        elif cmd in self.subcommands:
+            if '-h' in args.argv or '--help' in args.argv:
+                cmd = self.subcommands[cmd](parent=self)
+                cmd.parser.print_help()
+                return
+        else:
             self.print_help()
             return
 
@@ -240,21 +299,40 @@ Commands:\n""".format(self))
         if args.stderr:
             self.stderr = args.stderr
 
-        # Make the config object, and configure logging (or not).
+        # if args say not to "init" then we make a sort of empty config
         if args.no_init:
-            config = make_config([], extend=False)
-        else:
+            self.config = make_config([], extend=False, versioning=False)
+
+        else: # otherwise we make a proper config, and maybe turn on versioning
             logging.basicConfig()
-            config = make_config(args.config_paths, extend=args.extend_config)
+            self.config = make_config(args.config_paths, extend=args.extend_config, versioning=False)
+            if args.versioning:
+                configure_versioning(self.config, force=True)
+            elif not args.no_versioning:
+                configure_versioning(self.config)
+
+        # instantiate the subcommand
+        subcmd = self.subcommands[cmd](self, self.config)
+
+        # figure out if/how subcommand should show progress
+        subcmd.show_progress = args.progress
+        subcmd.progress = None
+        if subcmd.show_progress:
+            if args.progress_socket:
+                host, port = args.progress_socket.split(':')
+                subcmd.progress = SocketProgress(host, int(port))
+            else:
+                subcmd.progress = ConsoleProgress
+
+        # maybe should be verbose
+        subcmd.verbose = args.verbose
 
-        # And finally, do something of real value...
-        cmd = self.subcommands[cmd](self, config)
-        cmd.show_progress = args.progress
-        cmd.progress = Progress if args.progress else None
         # TODO: make this default to something from config?
-        cmd.runas_username = args.runas or None
-        log.debug("running '{} {}' with args: {}".format(self.name, cmd.name, args.argv))
-        cmd._run(*(args.command + args.argv))
+        subcmd.runas_username = args.runas or None
+
+        # and finally, run the subcommand
+        log.debug("running '%s %s' with args: %s", self.name, subcmd.name, args.argv)
+        subcmd._run(*(args.command + args.argv))
 
 
 class Subcommand(object):
@@ -270,7 +348,7 @@ class Subcommand(object):
         self.stdout = getattr(parent, 'stdout', sys.stdout)
         self.stderr = getattr(parent, 'stderr', sys.stderr)
         self.show_progress = show_progress
-        self.progress = Progress if show_progress else None
+        self.progress = ConsoleProgress if show_progress else None
         self.parser = argparse.ArgumentParser(
             prog='{0} {1}'.format(getattr(self.parent, 'name', 'UNDEFINED'), self.name),
             description=self.description)
@@ -290,18 +368,49 @@ class Subcommand(object):
         pass
 
     def make_session(self):
-        return self.parent.db_session_factory()
+        session = self.parent.db_session_factory()
+        user = self.get_runas_user(session=session)
+        if user:
+            session.set_continuum_user(user)
+        return session
+
+    def finalize_session(self, session, dry_run=False, success=True):
+        """
+        Wrap up the given session, per the given arguments.  This is meant to
+        provide a simple convenience, for commands which must do work within a
+        DB session, but would like to support a "dry run" mode.
+        """
+        if dry_run:
+            session.rollback()
+            log.info("dry run, so transaction was aborted")
+        else:
+            session.commit()
+            log.info("transaction was committed")
+        session.close()
 
     def get_runas_user(self, session=None, username=None):
         """
         Returns a proper User object, which the app should "run as"
         """
+        from sqlalchemy.orm.exc import NoResultFound
         from rattail.db.util import short_session
 
         if username is None:
-            username = self.runas_username or self.config.require('rattail', 'runas.default')
-        with short_session(session) as s:
-            return s.query(self.model.User).filter_by(username=username).one()
+            if hasattr(self, 'runas_username'):
+                username = self.runas_username
+            if not username and self.config:
+                username = self.config.get('rattail', 'runas.default')
+        if username:
+            user = None
+            with short_session(session) as s:
+                try:
+                    user = s.query(self.model.User).filter_by(username=username).one()
+                except NoResultFound:
+                    pass
+                else:
+                    if not session:
+                        s.expunge(user)
+            return user
 
     def progress_loop(self, func, items, factory=None, **kwargs):
         return progress_loop(func, items, factory or self.progress, **kwargs)
@@ -354,11 +463,11 @@ class CloneDatabase(Subcommand):
         dst_session = Session(bind=target_engine)
 
         for clsname in classes:
-            log.info("cloning data for model: {0}".format(clsname))
+            log.info("cloning data for model: %s", clsname)
             cls = getattr(model, clsname)
             src_query = src_session.query(cls)
             count = src_query.count()
-            log.debug("found {0} {1} records to clone".format(count, clsname))
+            log.debug("found %d %s records to clone", count, clsname)
             if not count:
                 continue
 
@@ -396,8 +505,19 @@ class DataSync(Subcommand):
 
         start = subparsers.add_parser('start', help="Start service")
         start.set_defaults(subcommand='start')
+
         stop = subparsers.add_parser('stop', help="Stop service")
         stop.set_defaults(subcommand='stop')
+
+        check = subparsers.add_parser('check', help="(DEPRECATED) Check for stale (lingering) changes in queue")
+        check.set_defaults(subcommand='check')
+
+        check_queue = subparsers.add_parser('check-queue', help="Check for stale (lingering) changes in queue")
+        check_queue.set_defaults(subcommand='check-queue')
+
+        check_watchers = subparsers.add_parser('check-watchers', help="Check for dead watcher threads")
+        check_watchers.set_defaults(subcommand='check-watchers')
+
         wait = subparsers.add_parser('wait', help="Wait for changes to be processed")
         wait.set_defaults(subcommand='wait')
 
@@ -410,9 +530,11 @@ class DataSync(Subcommand):
                             action='store_false', dest='daemonize',
                             help="Do not daemonize when starting.")
         parser.add_argument('-T', '--timeout', metavar='MINUTES', type=int, default=0,
-                            help="Optional timeout (in minutes) for use with the 'wait' command.  "
-                            "If specified, the waiting still stop after the given number of minutes "
-                            "and exit with a nonzero code to indicate failure.")
+                            help="Optional timeout (in minutes) for use with the 'wait' or 'check' commands.  "
+                            "If specified for 'wait', the waiting still stop after the given number of minutes "
+                            "and exit with a nonzero code to indicate failure.  If specified for 'check', the "
+                            "command will perform some health check based on the given timeout, and exit with "
+                            "nonzero code if the check fails.")
 
     def run(self, args):
         from rattail.datasync.daemon import DataSyncDaemon
@@ -420,6 +542,15 @@ class DataSync(Subcommand):
         if args.subcommand == 'wait':
             self.wait(args)
 
+        elif args.subcommand == 'check':
+            self.check_queue(args)
+
+        elif args.subcommand == 'check-queue':
+            self.check_queue(args)
+
+        elif args.subcommand == 'check-watchers':
+            self.check_watchers(args)
+
         else: # manage the daemon
             daemon = DataSyncDaemon(args.pidfile, config=self.config)
             if args.subcommand == 'stop':
@@ -434,26 +565,26 @@ class DataSync(Subcommand):
                         raise
 
     def wait(self, args):
-        model = self.parent.db_model
-        Session = self.parent.db_session_factory
-        session = Session()
-        started = datetime.datetime.utcnow()
+        model = self.model
+        session = self.make_session()
+        started = make_utc()
         log.debug("will wait for current change queue to clear")
         last_logged = started
 
         changes = session.query(model.DataSyncChange)
         count = changes.count()
-        log.debug("there are {} changes in the queue".format(count))
+        log.debug("there are %d changes in the queue", count)
         while count:
             try:
-                now = datetime.datetime.utcnow()
+                now = make_utc()
 
                 if args.timeout and (now - started).seconds >= (args.timeout * 60):
-                    log.warning("datasync wait timed out after {} minutes, with {} changes in queue".format(args.timeout, count))
+                    log.warning("datasync wait timed out after %d minutes, with %d changes in queue",
+                                args.timeout, count)
                     sys.exit(1)
 
                 if (now - last_logged).seconds >= 60:
-                    log.debug("still waiting, {} changes in the datasync queue".format(count))
+                    log.debug("still waiting, %d changes in the datasync queue", count)
                     last_logged = now
 
                 time.sleep(1)
@@ -467,6 +598,61 @@ class DataSync(Subcommand):
         session.close()
         log.debug("all datasync changes have been processed")
 
+    def check_queue(self, args):
+        """
+        Perform general queue / health check for datasync.
+        """
+        model = self.model
+        session = self.make_session()
+
+        # looking for changes which have been around for "timeout" minutes
+        timeout = args.timeout or 90
+        cutoff = make_utc() - datetime.timedelta(seconds=60 * timeout)
+        changes = session.query(model.DataSyncChange)\
+                         .filter(model.DataSyncChange.obtained < cutoff)\
+                         .count()
+        session.close()
+
+        # if we found stale changes, then "fail" - otherwise we'll "succeed"
+        if changes:
+            # TODO: should log set of unique payload types, for troubleshooting
+            log.debug("found %s changes, in queue for %s minutes", changes, timeout)
+            self.stderr.write("Found {} changes, in queue for {} minutes\n".format(changes, timeout))
+            sys.exit(1)
+
+        log.info("found no changes in queue for %s minutes", timeout)
+
+    def check_watchers(self, args):
+        """
+        Perform general health check for datasync watcher threads.
+        """
+        from rattail.datasync.config import load_profiles
+        from rattail.datasync.util import get_lastrun
+
+        profiles = load_profiles(self.config)
+        session = self.make_session()
+
+        # cutoff is "timeout" minutes before "now"
+        timeout = args.timeout or 15
+        cutoff = make_utc() - datetime.timedelta(seconds=60 * timeout)
+
+        dead = []
+        for key in profiles:
+
+            # looking for watcher "last run" time older than "timeout" minutes
+            lastrun = get_lastrun(self.config, key, tzinfo=False, session=session)
+            if lastrun and lastrun < cutoff:
+                dead.append(key)
+
+        session.close()
+
+        # if we found dead watchers, then "fail" - otherwise we'll "succeed"
+        if dead:
+            self.stderr.write("Found {} watcher threads dead for {} minutes: {}\n".format(len(dead), timeout, ', '.join(dead)))
+            sys.exit(1)
+
+        log.info("found no watcher threads dead for %s minutes", timeout)
+
 
 class EmailBouncer(Subcommand):
     """
@@ -616,7 +802,7 @@ class Dump(Subcommand):
 
         progress = None
         if self.show_progress: # pragma no cover
-            progress = Progress
+            progress = ConsoleProgress
 
         if args.output:
             output = open(args.output, 'wb')
@@ -663,7 +849,7 @@ class FileMonitorCommand(Subcommand):
         stop = subparsers.add_parser('stop', help="Stop service")
         stop.set_defaults(subcommand='stop')
 
-        if sys.platform == 'linux2':
+        if sys.platform in ('linux', 'linux2'):
             parser.add_argument('-p', '--pidfile',
                                 help="Path to PID file.", metavar='PATH')
             parser.add_argument('--daemonize', action='store_true', default=True, # TODO: should default to False
@@ -689,7 +875,7 @@ class FileMonitorCommand(Subcommand):
             uninstall.set_defaults(subcommand='remove')
 
     def run(self, args):
-        if sys.platform == 'linux2':
+        if sys.platform in ('linux', 'linux2'):
             from rattail.filemon import linux as filemon
 
             if args.subcommand == 'start':
@@ -728,7 +914,7 @@ class FileMonitorCommand(Subcommand):
 
                 password = ''
                 while password == '':
-                    password = getpass("Password for service user: ").strip()
+                    password = getpass(str("Password for service user: ")).strip()
                 options.extend(['--username', r'{0}\{1}'.format(server, username)])
                 options.extend(['--password', password])
 
@@ -771,7 +957,7 @@ class LoadHostDataCommand(Subcommand):
             sys.exit(1)
 
         proc = load.LoadProcessor(self.config)
-        proc.load_all_data(engines['host'], Progress)
+        proc.load_all_data(engines['host'], ConsoleProgress)
 
 
 class MakeAppDir(Subcommand):
@@ -835,12 +1021,20 @@ class MakeConfig(Subcommand):
                             "case the filename is inferred from 'type'.  Default is to current "
                             "working folder.")
 
-    def run(self, args):
+    def find_template(self, name):
         from rattail.files import resource_path
 
-        template_path = resource_path('rattail:data/config/{}.conf'.format(args.type))
-        if not os.path.exists(template_path):
-            self.stderr.write("config template not found: {}\n".format(template_path))
+        template_paths = self.config.getlist('rattail.config', 'templates',
+                                             default=['rattail:data/config'])
+        for template_path in template_paths:
+            path = resource_path('{}/{}.conf'.format(template_path.rstrip('/'), name))
+            if os.path.exists(path):
+                return path
+
+    def run(self, args):
+        template_path = self.find_template(args.type)
+        if not template_path:
+            self.stderr.write("config template not found for type: {}\n".format(args.type))
             sys.exit(1)
 
         output_path = os.path.abspath(args.output)
@@ -873,6 +1067,9 @@ class MakeUser(Subcommand):
                             help="Full (display) name for the new user (if applicable).")
         parser.add_argument('--comment',
                             help="Comment string for the new user (if applicable).")
+        parser.add_argument('--groups',
+                            help="Optional list of groups (roles) to which the new user "
+                            "should be assigned.")
 
     def run(self, args):
         mkuser = getattr(self, 'mkuser_{}'.format(args.system), None)
@@ -895,31 +1092,51 @@ class MakeUser(Subcommand):
         try:
             password = None
             while not password:
-                password = getpass(b"enter password for new user: ").strip()
+                password = getpass(str("enter password for new user: ")).strip()
         except KeyboardInterrupt:
             self.stderr.write("\noperation canceled by user\n")
             sys.exit(2)
         return password
 
     def mkuser_rattail(self, args):
+        from sqlalchemy.orm.exc import NoResultFound
         from rattail.db import auth
 
-        session = self.parent.db_session_factory()
+        session = self.make_session()
         model = self.parent.db_model
         if session.query(model.User).filter_by(username=args.username).count():
             session.close()
             return self.user_exists(args)
 
+        roles = []
+        if args.groups:
+            for name in parse_list(args.groups):
+                try:
+                    role = session.query(model.Role)\
+                                  .filter(model.Role.name == name)\
+                                  .one()
+                except NoResultFound:
+                    self.stderr.write("Role not found: {}\n".format(name))
+                    session.close()
+                    sys.exit(4)
+                else:
+                    roles.append(role)
+
         user = model.User(username=args.username)
         auth.set_user_password(user, self.obtain_password(args))
+
         if args.admin:
             user.roles.append(auth.administrator_role(session))
+        for role in roles:
+            user.roles.append(role)
+
         if args.full_name:
             kwargs = {'display_name': args.full_name}
             words = args.full_name.split()
             if len(words) == 2:
                 kwargs.update({'first_name': words[0], 'last_name': words[1]})
             user.person = model.Person(**kwargs)
+
         session.add(user)
         session.commit()
         session.close()
@@ -984,16 +1201,222 @@ class PalmCommand(Subcommand):
         if args.subcommand == 'register':
             try:
                 palm.register_conduit()
-            except PalmError, error:
-                sys.stderr.write(str(error) + '\n')
+            except PalmError as error:
+                sys.stderr.write("{}\n".format(error))
 
         elif args.subcommand == 'unregister':
             try:
                 palm.unregister_conduit()
-            except PalmError, error:
-                sys.stderr.write(str(error) + '\n')
+            except PalmError as error:
+                sys.stderr.write("{}\n".format(error))
                 
 
+class RunAndMail(Subcommand):
+    """
+    Run a command as subprocess, and email the result/output
+    """
+    name = 'run-n-mail'
+    description = __doc__.strip()
+
+    def add_parser_args(self, parser):
+        parser.add_argument('--key', default='run_n_mail',
+                            help="Config key for email settings")
+        # TODO: these all seem like good ideas, but not needed yet?
+        # parser.add_argument('--from', '-F', metavar='ADDRESS',
+        #                     help="Override value of From: header")
+        # parser.add_argument('--to', '-T', metavar='ADDRESS',
+        #                     help="Override value of To: header (may specify more than once)")
+        # parser.add_argument('--cc', metavar='ADDRESS',
+        #                     help="Override value of Cc: header (may specify more than once)")
+        # parser.add_argument('--bcc', metavar='ADDRESS',
+        #                     help="Override value of Bcc: header (may specify more than once)")
+        parser.add_argument('--subject', '-S',
+                            help="Override value of Subject: header (i.e. value after prefix)")
+        parser.add_argument('cmd', metavar='COMMAND',
+                            help="Command which should be ran, and result of which will be emailed")
+
+    def run(self, args):
+        from rattail.mail import send_email
+
+        cmd = parse_list(args.cmd)
+        log.info("will run command as subprocess: %s", cmd)
+        run_began = make_utc()
+
+        try:
+            # TODO: must we allow for shell=True in some situations? (clearly not yet)
+            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
+            retcode = 0
+            log.info("command completed successfully")
+        except subprocess.CalledProcessError as error:
+            output = error.output
+            retcode = error.returncode
+            log.warn("command exited with code: %s", retcode)
+
+        if six.PY3:
+            output = output.decode('ascii')
+
+        run_ended = make_utc()
+        runtime = run_ended - run_began
+        runtime_pretty = humanize.naturaldelta(runtime)
+
+        kwargs = {}
+        if args.subject:
+            kwargs['subject_template'] = args.subject
+
+        send_email(self.config, args.key, {
+            'cmd': cmd,
+            'run_began': run_began,
+            'run_ended': run_ended,
+            'runtime': runtime,
+            'runtime_pretty': runtime_pretty,
+            'retcode': retcode,
+            'output': output,
+        }, **kwargs)
+
+
+class RunSQL(Subcommand):
+    """
+    Run (first statement of) a SQL script against a database
+    """
+    name = 'runsql'
+    description = __doc__.strip()
+
+    def add_parser_args(self, parser):
+        parser.add_argument('engine',
+                            help="SQLAlchemy engine URL for the database.")
+        parser.add_argument('script', type=argparse.FileType('r'),
+                            help="Path to file which contains a SQL script.")
+        parser.add_argument('--max-width', type=int, default=80,
+                            help="Max table width when displaying results.")
+
+    def run(self, args):
+        import sqlalchemy as sa
+        import texttable
+
+        sql = []
+        for line in args.script:
+            line = line.strip()
+            if line and not line.startswith('--'):
+                sql.append(line)
+                if line.endswith(';'):
+                    break
+
+        sql = ' '.join(sql)
+        engine = sa.create_engine(args.engine)
+
+        result = engine.execute(sql)
+        rows = result.fetchall()
+        if rows:
+            table = texttable.Texttable(max_width=args.max_width)
+
+            # force all columns to be treated as text.  that seems a bit
+            # heavy-handed but is definitely the simplest way to deal with
+            # issues such as a UPC being displayed in scientific notation
+            table.set_cols_dtype(['t' for col in rows[0]])
+
+            # add a header row, plus all data rows
+            table.add_rows([rows[0].keys()] + rows)
+
+            self.stdout.write("{}\n".format(table.draw()))
+
+
+class Upgrade(Subcommand):
+    """
+    Upgrade the local Rattail app
+    """
+    name = 'upgrade'
+    description = __doc__.strip()
+
+    def add_parser_args(self, parser):
+        parser.add_argument('--description',
+                            help="Description for the new/matched upgrade.")
+        parser.add_argument('--enabled', action='store_true', default=True,
+                            help="Indicate the enabled flag should be ON for the new/matched upgrade.  "
+                            "Note that this is the default if you do not specify.")
+        parser.add_argument('--no-enabled', action='store_false', dest='enabled',
+                            help="Indicate the enabled flag should be OFF for the new/matched upgrade.")
+        parser.add_argument('--create', action='store_true',
+                            help="Create a new upgrade with the given attributes.")
+        parser.add_argument('--execute', action='store_true',
+                            help="Execute the upgrade.  Note that if you do not specify "
+                            "--create then the upgrade matching the given attributes "
+                            "will be read from the database.  If such an upgrade is not "
+                            "found or is otherwise invalid (e.g. already executed), "
+                            "the command will fail.")
+        parser.add_argument('--dry-run', action='store_true',
+                            help="Go through the full motions and allow logging etc. to "
+                            "occur, but rollback (abort) the transaction at the end.")
+
+    def run(self, args):
+        from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
+        from rattail.upgrades import get_upgrade_handler
+
+        if not args.create and not args.execute:
+            self.stderr.write("Must specify --create and/or --execute\n")
+            sys.exit(1)
+
+        session = self.make_session()
+        model = self.model
+        user = self.get_runas_user(session)
+
+        if args.create:
+            upgrade = model.Upgrade()
+            upgrade.description = args.description
+            upgrade.created = make_utc()
+            upgrade.created_by = user
+            upgrade.enabled = args.enabled
+            session.add(upgrade)
+            session.flush()
+            log.info("user '%s' created new upgrade: %s", user.username, upgrade)
+
+        else:
+            upgrades = session.query(model.Upgrade)\
+                              .filter(model.Upgrade.enabled == args.enabled)
+            if args.description:
+                upgrades = upgrades.filter(model.Upgrade.description == args.description)
+            try:
+                upgrade = upgrades.one()
+            except NoResultFound:
+                self.stderr.write("no matching upgrade found\n")
+                session.rollback()
+                session.close()
+                sys.exit(1)
+            except MultipleResultsFound:
+                self.stderr.write("found {} matching upgrades\n".format(upgrades.count()))
+                session.rollback()
+                session.close()
+                sys.exit(1)
+
+        if args.execute:
+            if upgrade.executed:
+                self.stderr.write("upgrade has already been executed: {}\n".format(upgrade))
+                session.rollback()
+                session.close()
+                sys.exit(1)
+            if not upgrade.enabled:
+                self.stderr.write("upgrade is not enabled for execution: {}\n".format(upgrade))
+                session.rollback()
+                session.close()
+                sys.exit(1)
+
+            # execute upgrade
+            handler = get_upgrade_handler(self.config)
+            log.info("will now execute upgrade: %s", upgrade)
+            if not args.dry_run:
+                handler.mark_executing(upgrade)
+                session.commit()
+                handler.do_execute(upgrade, user, progress=self.progress)
+            log.info("user '%s' executed upgrade: %s", user.username, upgrade)
+
+        if args.dry_run:
+            session.rollback()
+            log.info("dry run, so transaction was rolled back")
+        else:
+            session.commit()
+            log.info("transaction was committed")
+        session.close()
+
+
 def main(*args):
     """
     The primary entry point for the Rattail command system.
