@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -31,15 +31,18 @@ import re
 import sys
 import shlex
 import datetime
-import ConfigParser
+from six.moves import configparser
 import warnings
 import logging
 import logging.config
 
+import six
+
 from rattail.util import load_entry_points, import_module_path
 from rattail.exceptions import WindowsExtensionsNotInstalled, ConfigurationError
 from rattail.files import temp_path
 from rattail.logging import TimeConverter
+from rattail.util import prettify
 
 
 log = logging.getLogger(__name__)
@@ -67,14 +70,16 @@ def parse_list(value):
         return []
     # Per the shlex docs (https://docs.python.org/2/library/shlex.html):
     # "Prior to Python 2.7.3, this module did not support Unicode input."
-    if isinstance(value, unicode):
+    if six.PY2 and isinstance(value, six.text_type):
         value = value.encode('utf-8')
     parser = shlex.shlex(value)
-    parser.whitespace += u','
+    parser.whitespace += ','
     parser.whitespace_split = True
     values = list(parser)
     for i, value in enumerate(values):
-        if value.startswith(u'"') and value.endswith(u'"'):
+        if value.startswith('"') and value.endswith('"'):
+            values[i] = value[1:-1]
+        elif value.startswith("'") and value.endswith("'"):
             values[i] = value[1:-1]
     return values
 
@@ -108,7 +113,10 @@ class RattailConfig(object):
     def __init__(self, files=[], usedb=None, preferdb=None):
         self.files_requested = []
         self.files_read = []
-        self.parser = ConfigParser.SafeConfigParser()
+        if six.PY3:
+            self.parser = configparser.ConfigParser()
+        else: # PY2
+            self.parser = configparser.SafeConfigParser()
         for path in files:
             self.read_file(path)
         self.usedb = usedb
@@ -146,7 +154,11 @@ class RattailConfig(object):
         log.debug("will attempt to read config from file: {0}".format(path))
         self.files_requested.append(path)
 
-        parser = ConfigParser.SafeConfigParser(dict(
+        if six.PY3:
+            ConfigParser = configparser.ConfigParser
+        else: # PY2
+            ConfigParser = configparser.SafeConfigParser
+        parser = ConfigParser(dict(
             here=os.path.dirname(path),
         ))
         if not parser.read(path):
@@ -189,7 +201,7 @@ class RattailConfig(object):
 
         try:
             logging.config.fileConfig(path, disable_existing_loggers=False)
-        except ConfigParser.NoSectionError as error:
+        except configparser.NoSectionError as error:
             log.warning("tried to configure logging, but got NoSectionError: {0}".format(error))
         else:
             log.debug("configured logging")
@@ -311,7 +323,7 @@ class RattailConfig(object):
         value = self.get(*args, **kwargs)
         if value is None:
             return None
-        if isinstance(value, basestring):
+        if isinstance(value, six.string_types):
             return parse_list(value)
         return value            # maybe a caller-provided default?
 
@@ -352,6 +364,35 @@ class RattailConfig(object):
     # convenience methods
     ##############################
 
+    def app_title(self, default=None):
+        """
+        Returns official display title for the current app.
+        """
+        if not default:
+            return self.require('rattail', 'app_title')
+        return self.get('rattail', 'app_title', default=default)
+
+    def node_title(self, default=None):
+        """
+        Returns official display title for the current app.
+        """
+        title = self.get('rattail', 'node_title')
+        if title:
+            return title
+        return self.app_title(default=default)
+
+    def node_type(self, default=None):
+        """
+        Returns the "type" of current node.  What this means will generally
+        depend on the app logic.
+        """
+        try:
+            return self.require('rattail', 'node_type', usedb=False)
+        except ConfigurationError:
+            if default:
+                return default
+            raise
+
     def production(self):
         """
         Returns boolean indicating whether the app is running in production mode
@@ -371,9 +412,23 @@ class RattailConfig(object):
         return self.getbool('rattail.db', 'versioning.enabled', usedb=False,
                             default=False)
 
-    def workdir(self, require=False):
+    def appdir(self, require=True):
         """
-        Returns boolean indicating whether the config indicates production mode.
+        Returns path to the 'app' dir, if known.
+        """
+        get = self.require if require else self.get
+        return get('rattail', 'appdir')
+
+    def datadir(self, require=True):
+        """
+        Returns path to the 'data' dir, if known.
+        """
+        get = self.require if require else self.get
+        return get('rattail', 'datadir')
+
+    def workdir(self, require=True):
+        """
+        Returns path to the 'work' dir, if known.
         """
         get = self.require if require else self.get
         return get('rattail', 'workdir')
@@ -424,20 +479,94 @@ class RattailConfig(object):
             return os.path.join(filedir, filename)
         return filedir
 
+    def upgrade_filedir(self):
+        """
+        Returns path to root folder where upgrade files are stored.
+        """
+        path = os.path.abspath(self.require('rattail.upgrades', 'files'))
+        return path
+
+    def upgrade_filepath(self, uuid, filename=None, makedirs=False):
+        """
+        Returns absolute path to upgrade data file, generated from the given args.
+        """
+        rootdir = self.upgrade_filedir()
+        filedir = os.path.join(rootdir, uuid[:2], uuid[2:])
+        if makedirs and not os.path.exists(filedir):
+            os.makedirs(filedir)
+        if filename:
+            return os.path.join(filedir, filename)
+        return filedir
+
+    def upgrade_command(self, default='/bin/sleep 30'):
+        """
+        Returns command to be used when performing upgrades.
+        """
+        # NOTE: we don't allow command to be specified in DB, for security reasons..
+        return self.getlist('rattail.upgrades', 'command', default=default, usedb=False)
+
     def datasync_url(self):
         """
         Returns configured URL for managing datasync daemon.
         """
         return self.get('rattail.datasync', 'url')
 
-    def get_enum(self):
+    def get_enum(self, **kwargs):
         """
         Returns a reference to configured 'enum' module; defaults to
         :mod:`rattail.enum`.
         """
-        spec = self.get('rattail', 'enum', default='rattail.enum')
+        kwargs.setdefault('usedb', False)
+        spec = self.get('rattail', 'enum', default='rattail.enum', **kwargs)
         return import_module_path(spec)
 
+    def get_model(self):
+        """
+        Returns a reference to configured 'model' module; defaults to
+        :mod:`rattail.db.model`.
+        """
+        spec = self.get('rattail', 'model', default='rattail.db.model', usedb=False)
+        return import_module_path(spec)
+
+    def get_trainwreck_model(self):
+        """
+        Returns a reference to the configured data 'model' module for
+        Trainwreck.  Note that there is *not* a default value for this; it must
+        be configured.
+        """
+        spec = self.require('rattail.trainwreck', 'model', usedb=False)
+        return import_module_path(spec)
+
+    def product_key(self):
+        """
+        Returns the name of the attribute which should be treated as the
+        canonical product key field, e.g. 'upc' or 'item_id' etc.
+        """
+        return self.get('rattail', 'product.key', default='upc')
+
+    def product_key_title(self, key=None):
+        """
+        Returns the title string to be used when displaying product key field,
+        e.g. "UPC" or "Part No." etc.
+        """
+        title = self.get('rattail', 'product.key_title')
+        if title:
+            return title
+        if not key:
+            key = self.product_key()
+        if key == 'upc':
+            return "UPC"
+        if key == 'item_id':
+            return "Item ID"
+        return prettify(key)
+
+    def single_store(self):
+        """
+        Returns boolean indicating whether the system is configured to behave
+        as if it belongs to a single Store.
+        """
+        return self.getbool('rattail', 'single_store', default=False)
+
     def get_store(self, session):
         """
         Returns a :class:`rattail.db.model.Store` instance corresponding to app
@@ -483,18 +612,26 @@ class ConfigExtension(object):
         """
 
 
-def make_config(files=None, usedb=None, preferdb=None, env=os.environ, winsvc=None, extend=True):
+def make_config(files=None, usedb=None, preferdb=None, env=None, winsvc=None, extend=True, versioning=None):
     """
     Returns a new config object, initialized with the given parameters and
     further modified by all registered config extensions.
+
+    :param versioning: Controls whether or not the versioning system is
+       configured with the new config object.  If ``True``, versioning will be
+       configured.  If ``False`` then it will not be configured.  If ``None``
+       (the default) then versioning will be configured only if the config
+       object itself says that it should be.
     """
+    if env is None:
+        env = os.environ
     if files is None:
         files = env.get('RATTAIL_CONFIG_FILES')
         if files is not None:
             files = files.split(os.pathsep)
         else:
             files = default_system_paths() + default_user_paths()
-    elif isinstance(files, basestring):
+    elif isinstance(files, six.string_types):
         files = [files]
 
     # If making config for a Windows service, we must read the default config
@@ -504,7 +641,7 @@ def make_config(files=None, usedb=None, preferdb=None, env=os.environ, winsvc=No
     # pass parameters to a Windows service.  This way we can effectively do
     # just that, via config.
     if winsvc is not None:
-        parser = ConfigParser.SafeConfigParser()
+        parser = configparser.SafeConfigParser()
         parser.read(files)
         if parser.has_section('rattail.config'):
             key = 'winsvc.{0}'.format(winsvc)
@@ -516,13 +653,33 @@ def make_config(files=None, usedb=None, preferdb=None, env=os.environ, winsvc=No
     config = RattailConfig(files, usedb=usedb, preferdb=preferdb)
     config.configure_logging()
 
+    if config.getbool('rattail', 'suppress_psycopg2_wheel_warning', usedb=False):
+        # TODO: revisit this, does it require action from us?
+        # suppress this warning about psycopg2 wheel; not sure what it means yet
+        # exactly but it's causing frequent noise for us...
+        warnings.filterwarnings(
+            'ignore',
+            r'^The psycopg2 wheel package will be renamed from release 2\.8; in order to keep '
+            'installing from binary please use "pip install psycopg2-binary" instead\. For details '
+            'see: <http://initd.org/psycopg/docs/install.html#binary-install-from-pypi>\.',
+            UserWarning,
+            r'^psycopg2$',
+        )
+
     # Apply extra config for all available extensions.
     if extend:
         extensions = load_entry_points('rattail.config.extensions')
-        for extension in extensions.itervalues():
+        for extension in extensions.values():
             log.debug("applying '{0}' config extension".format(extension.key))
             extension().configure(config)
 
+    # maybe configure versioning
+    if versioning is None:
+        versioning = config.versioning_enabled()
+    if versioning:
+        from rattail.db.config import configure_versioning
+        configure_versioning(config)
+
     return config
 
 
@@ -644,17 +801,29 @@ class ConfigProfile(object):
                                    default=default)
 
     def _config_int(self, option, minimum=1, default=None):
-        option = '{0}.{1}'.format(self.key, option)
-        if self.config.has_option(self.section, option):
-            value = self.config.getint(self.section, option)
+        """
+        Retrieve the *integer* value for the given option.
+        """
+        option = '{}.{}'.format(self.key, option)
+
+        # try to read value from config
+        value = self.config.getint(self.section, option)
+        if value is not None:
+
+            # found a value; validate it
             if value < minimum:
-                log.warning("config value {0} is too small; falling back to minimum "
-                            "of {1} for option: {2}".format(value, minimum, option))
+                log.warning("config value %s is too small; falling back to minimum "
+                            "of %s for option: %s", value, minimum, option)
                 value = minimum
+
+        # or, use default value, if valid
         elif default is not None and default >= minimum:
             value = default
+
+        # or, just use minimum value
         else:
             value = minimum
+
         return value
 
     def _config_list(self, option):
@@ -678,7 +847,7 @@ class FreeTDSLoggingFilter(logging.Filter):
             and record.levelno == logging.ERROR
             and record.msg == "Exception during reset or similar"
             and record.exc_info
-            and self.pattern.search(str(record.exc_info[1]))):
+            and self.pattern.search(six.text_type(record.exc_info[1]))):
 
             # Log this as a warning instead of error, to cut down on our noise.
             record.levelno = logging.WARNING
