@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2019 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -52,7 +52,8 @@ class DataSyncProfile(ConfigProfile):
         if self.watcher_spec == 'null':
             self.watcher = NullWatcher(config, key)
         else:
-            self.watcher = load_object(self.watcher_spec)(config, key)
+            dbkey = self._config_string('watcher.db', default='default')
+            self.watcher = load_object(self.watcher_spec)(config, key, dbkey=dbkey)
         self.watcher.delay = self._config_int('watcher.delay', default=self.watcher.delay)
         self.watcher.retry_attempts = self._config_int('watcher.retry_attempts', default=self.watcher.retry_attempts)
         self.watcher.retry_delay = self._config_int('watcher.retry_delay', default=self.watcher.retry_delay)
@@ -62,34 +63,33 @@ class DataSyncProfile(ConfigProfile):
             self.watcher.consumes_self = True
         else:
             self.watcher.consumes_self = False
-            self.consumers = self.normalize_consumers(consumers)
             self.consumer_delay = self._config_int('consumer_delay', default=1)
-            self.watcher.consumer_stub_keys = [c.key for c in self.isolated_consumers]
-            if self.common_consumers:
-                self.watcher.consumer_stub_keys.append(None)
-
-    @property
-    def isolated_consumers(self):
-        return [c for c in self.consumers if c.isolated]
-
-    @property
-    def common_consumers(self):
-        return [c for c in self.consumers if not c.isolated]
+            self.consumers = self.normalize_consumers(consumers)
+            self.watcher.consumer_stub_keys = [c.key for c in self.consumers]
 
     def normalize_consumers(self, raw_consumers):
         # TODO: This will do for now, but isn't as awesome as it probably could be.
-        default_isolated = self._config_boolean('consumers.isolated', default=True)
         default_runas = self._config_string('consumers.runas')
         consumers = []
         for key in raw_consumers:
             consumer_spec = self._config_string('consumer.{}'.format(key))
+            if consumer_spec == 'null':
+                consumer_spec = 'rattail.datasync.consumers:NullTestConsumer'
             dbkey = self._config_string('consumer.{}.db'.format(key), default=key)
             runas = self._config_string('consumer.{}.runas'.format(key))
             consumer = load_object(consumer_spec)(self.config, key, dbkey=dbkey,
-                                                  runas=runas or default_runas)
+                                                  runas=runas or default_runas,
+                                                  watcher=self.watcher)
             consumer.spec = consumer_spec
-            consumer.isolated = self._config_boolean('consumer.{}.isolated'.format(key),
-                                                     default=default_isolated)
+            consumer.delay = self._config_int(
+                'consumer.{}.delay'.format(key),
+                default=self.consumer_delay)
+            consumer.retry_attempts = self._config_int(
+                'consumer.{}.retry_attempts'.format(key),
+                default=consumer.retry_attempts)
+            consumer.retry_delay = self._config_int(
+                'consumer.{}.retry_delay'.format(key),
+                default=consumer.retry_delay)
             consumers.append(consumer)
         return consumers
 
