@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2018 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -26,6 +26,10 @@ DataSync Watchers
 
 from __future__ import unicode_literals, absolute_import
 
+import datetime
+
+from rattail.time import make_utc, localtime
+
 
 class DataSyncWatcher(object):
     """
@@ -35,9 +39,10 @@ class DataSyncWatcher(object):
     retry_attempts = 1
     retry_delay = 1 # seconds
 
-    def __init__(self, config, key):
+    def __init__(self, config, key, dbkey=None):
         self.config = config
         self.key = key
+        self.dbkey = dbkey
         self.delay = 1 # seconds
 
     def setup(self):
@@ -45,6 +50,38 @@ class DataSyncWatcher(object):
         This method is called when the watcher thread is first started.
         """
 
+    def localize_lastrun(self, session, lastrun):
+        """
+        Calculates a timestamp using `lastrun` as the starting point, but also
+        taking into account a possible (?) time drift between the local and
+        "other" server.
+        """
+        # get current time according to "other" server, and convert to UTC
+        before = make_utc()
+        other_now = session.execute('select current_timestamp').fetchone()[0]
+        after = make_utc()
+        other_now = make_utc(localtime(self.config, other_now))
+
+        # get current time according to local server (i.e. Rattail), in UTC
+        local_now = after
+
+        # drift is essentially the difference between the two timestamps.  it
+        # is meant to be a positive value (or zero) since it will be
+        # "subtracted" from the `lastrun` time in order to obtain the timestamp
+        # we should use for "other" queries.  note that we also add one second
+        # to the drift, just to be on the safe side.
+        if other_now < local_now:
+            drift = local_now - other_now
+        else:
+            drift = datetime.timedelta(seconds=0)
+        drift += datetime.timedelta(seconds=1)
+
+        # convert lastrun time to local timezone, for "other" queries
+        lastrun = localtime(self.config, lastrun, tzinfo=False)
+
+        # and finally, apply the drift
+        return lastrun - drift
+
     def get_changes(self, lastrun):
         """
         This must be implemented by the subclass.  It should check the source
