@@ -62,6 +62,11 @@ class PricingBatch(BatchMixin, Base):
     Base name of the input data file, if applicable.
     """)
 
+    shelved = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
+    Flag to indicate whether the batch is "shelved" for later processing.  This
+    may be used to assist with workflow when entering/executing new batches.
+    """)
+
 
 class PricingBatchRow(ProductBatchRowMixin, Base):
     """
