@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2019 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -39,6 +39,9 @@ class PhoneNumber(Base):
     Represents a phone (or fax) number associated with a contactable entity.
     """
     __tablename__ = 'phone'
+    __table_args__ = (
+        sa.Index('phone_ix_parent', 'parent_type', 'parent_uuid'),
+    )
     __versioned__= {}
 
     uuid = uuid_column()
@@ -60,6 +63,9 @@ class EmailAddress(Base):
     Represents an email address associated with a contactable entity.
     """
     __tablename__ = 'email'
+    __table_args__ = (
+        sa.Index('email_ix_parent', 'parent_type', 'parent_uuid'),
+    )
     __versioned__= {}
 
     uuid = uuid_column()
@@ -86,6 +92,9 @@ class MailingAddress(Base):
     Represents a physical / mailing address associated with a contactable entity.
     """
     __tablename__ = 'address'
+    __table_args__ = (
+        sa.Index('address_ix_parent', 'parent_type', 'parent_uuid'),
+    )
     __versioned__= {}
 
     uuid = uuid_column()
