@@ -1,8 +1,8 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2017 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -24,8 +24,11 @@
 Data Models for Employees
 """
 
-from __future__ import unicode_literals
+from __future__ import unicode_literals, absolute_import
 
+import datetime
+
+import six
 import sqlalchemy as sa
 from sqlalchemy import orm
 from sqlalchemy.ext.associationproxy import association_proxy
@@ -50,10 +53,25 @@ class Employee(Base):
     __versioned__ = {}
 
     uuid = uuid_column()
-    id = sa.Column(sa.Integer())
+
+    id = sa.Column(sa.Integer(), nullable=True, doc="""
+    Numeric ID for the employee, if applicable/known.
+    """)
+
     person_uuid = sa.Column(sa.String(length=32), nullable=False)
-    status = sa.Column(sa.Integer())
-    display_name = sa.Column(sa.String(length=100))
+    person = orm.relationship(Person, back_populates='employee', doc="""
+    Reference to the person to whom this employee record pertains.
+    """)
+
+    status = sa.Column(sa.Integer(), nullable=True, doc="""
+    Status code for the employee, e.g. current vs. former.
+    """)
+
+    display_name = sa.Column(sa.String(length=100), nullable=True, doc="""
+    Display name for the employee.  Whereas the *person's* display name is
+    generally their full name, this *employee* display name is often an
+    abbreviated version which is visible on printed receipts, etc.
+    """)
 
     full_time = sa.Column(sa.Boolean(), nullable=True, doc="""
     Simple flag to indicate if employee is considered full-time.  This flag
@@ -64,8 +82,6 @@ class Employee(Base):
     Date on which employee became full time, if known/relevant.
     """)
 
-    person = orm.relationship(Person, back_populates='employee')
-
     first_name = association_proxy('person', 'first_name',
                                    getset_factory=getset_factory,
                                    creator=lambda n: Person(first_name=n))
@@ -74,13 +90,25 @@ class Employee(Base):
                                   creator=lambda n: Person(last_name=n))
     user = association_proxy('person', 'user')
 
-    def __unicode__(self):
-        return unicode(self.person)
+    def __str__(self):
+        if six.PY2:
+            return unicode(self).encode('utf8')
+        return str(self.person or "")
+
+    if six.PY2:
+        def __unicode__(self):
+            return unicode(self.person or "")
 
     @property
     def customers(self):
         return self.person.customers
 
+    def only_customer(self):
+        """
+        Convenience method which invokes ``self.person.only_customer()``
+        """
+        return self.person.only_customer()
+
     def add_email_address(self, address, type='Home'):
         email = EmployeeEmailAddress(address=address, type=type)
         self.emails.append(email)
@@ -89,6 +117,26 @@ class Employee(Base):
         phone = EmployeePhoneNumber(number=number, type=type)
         self.phones.append(phone)
 
+    def sorted_history(self, reverse=False):
+        """
+        Returns the employee history, sorted in a "reasonable" way.
+        """
+        old_date = datetime.date(1900, 1, 1)
+        new_date = datetime.date.today() + datetime.timedelta(days=7)
+        return sorted(self.history,
+                      key=lambda h: (h.start_date or old_date, h.end_date or new_date),
+                      reverse=reverse)
+
+    def get_current_history(self):
+        """
+        Returns the "current" history record for the employee, if applicable.
+        Note that this history record is not necessarily "active" - it's just
+        the most recent.
+        """
+        if self.history:
+            history = self.sorted_history(reverse=True)
+            return history[0]
+
 
 Person.employee = orm.relationship(
     Employee,
@@ -221,3 +269,36 @@ Department.employees = association_proxy(
     '_employees', 'employee',
     creator=lambda x: EmployeeDepartment(employee=x),
     getset_factory=getset_factory)
+
+
+class EmployeeHistory(Base):
+    """
+    Represents a period of time for which a person was employed by the organization.
+    """
+    __tablename__ = 'employee_history'
+    __table_args__ = (
+        sa.ForeignKeyConstraint(['employee_uuid'], ['employee.uuid'], name='employee_history_fk_employee'),
+    )
+    __versioned__ = {}
+
+    uuid = uuid_column()
+
+    employee_uuid = sa.Column(sa.String(length=32), nullable=False)
+    employee = orm.relationship(
+        Employee, doc="""
+        Reference to the employee to which this history record pertains.
+        """,
+        backref=orm.backref(
+            'history',
+            order_by='EmployeeHistory.start_date',
+            doc="""
+            Sequence of history records for the employee.
+            """))
+
+    start_date = sa.Column(sa.Date(), nullable=False, doc="""
+    Date on which the employee became active in the organization.
+    """)
+
+    end_date = sa.Column(sa.Date(), nullable=True, doc="""
+    Date on which the employee became *inactive*.
+    """)
