@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2018 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -59,6 +59,10 @@ class Role(Base):
     means the role has no say in the timeout.
     """)
 
+    notes = sa.Column(sa.Text(), nullable=True, doc="""
+    Any arbitrary notes for the role.
+    """)
+
     def __str__(self):
         return self.name or ''
 
@@ -103,7 +107,8 @@ class User(Base):
     __table_args__ = (
         sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='user_fk_person'),
         sa.UniqueConstraint('username', name='user_uq_username'),
-        )
+        sa.Index('user_ix_person', 'person_uuid'),
+    )
     __versioned__ = {'exclude': ['password', 'salt', 'last_login']}
 
     uuid = uuid_column()
