@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2018 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -264,7 +264,7 @@ class Importer(object):
                 if diffs:
                     log.debug("fields '{}' differed for local data: {}, host data: {}".format(
                         ','.join(diffs), local_data, host_data))
-                    local_object = self.update_object(local_object, host_data, local_data)
+                    local_object = self.update_object(local_object, host_data, local_data=local_data)
                     updated.append((local_object, local_data, host_data))
                     if self.max_update and len(updated) >= self.max_update:
                         log.warning("max of {} *updated* records has been reached; stopping now".format(self.max_update))
@@ -449,11 +449,28 @@ class Importer(object):
         """
         return obj
 
+    def get_local_objects(self, host_data=None):
+        """
+        Fetch all raw objects from the local system.
+        """
+        raise NotImplementedError
+
     def cache_local_data(self, host_data=None):
         """
         Cache all raw objects and normalized data from the local system.
         """
-        raise NotImplementedError
+        objects = self.get_local_objects(host_data=host_data)
+        cached = {}
+
+        def cache(obj, i):
+            data = self.normalize_local_object(obj)
+            key = self.get_key(data)
+            cached[key] = {'object': obj, 'data': data}
+
+        self.progress_loop(cache, objects,
+                           message="Reading {} data from {}".format(
+                               self.model_name, self.handler.local_title))
+        return cached
 
     def cache_local_message(self):
         """
