@@ -2,7 +2,7 @@
 ################################################################################
 #
 #  Rattail -- Retail Software Framework
-#  Copyright © 2010-2019 Lance Edgar
+#  Copyright © 2010-2020 Lance Edgar
 #
 #  This file is part of Rattail.
 #
@@ -35,6 +35,7 @@ from sqlalchemy import orm
 from sqlalchemy.orm.exc import NoResultFound
 
 from rattail.db import model, auth
+from rattail.core import get_uuid
 from rattail.importing import ToSQLAlchemy, BatchImporter
 from rattail.db.util import maxlen, normalize_full_name, format_phone_number, normalize_phone_number
 from rattail.time import localtime, make_utc
@@ -96,12 +97,219 @@ class GlobalMixin(object):
         return query
 
 
-class PersonImporter(ToRattail):
+class ContactMixin(object):
+    """
+    Mixin for importers which target a "contact" sort of record, e.g. customers.
+    """
+    phone_fields = [
+        'phone_number',
+        'phone_number_2',
+    ]
+    address_fields = [
+        'address_street',
+        'address_street2',
+        'address_city',
+        'address_state',
+        'address_zipcode',
+        'address_invalid',
+    ]
+    email_fields = [
+        'email_address',
+    ]
+
+    @property
+    def supported_fields(self):
+        fields = super(ContactMixin, self).supported_fields
+        return (fields
+                + self.phone_fields
+                + self.address_fields
+                + self.email_fields)
+
+    def cache_query_options(self):
+        options = super(ContactMixin, self).cache_query_options()
+        if self.fields_active(self.phone_fields):
+            options.append(orm.joinedload(self.model_class.phones))
+        if self.fields_active(self.address_fields):
+            options.append(orm.joinedload(self.model_class.addresses))
+        if self.fields_active(self.email_fields):
+            options.append(orm.joinedload(self.model_class.emails))
+        return options
+
+    def normalize_local_object(self, obj):
+        data = super(ContactMixin, self).normalize_local_object(obj)
+
+        if self.fields_active(self.phone_fields):
+            # TODO: how to filter the phone list?
+            # phones = [phone for phone in obj.phones
+            #           if phone.type == 'Voice']
+            phones = obj.phones
+            if 'phone_number' in self.fields:
+                data['phone_number'] = phones[0].number if phones else None
+            if 'phone_number_2' in self.fields:
+                data['phone_number_2'] = phones[1].number if len(phones) > 1 else None
+
+        if 'email_address' in self.fields:
+            email = obj.emails[0] if obj.emails else None
+            data['email_address'] = email.address if email else None
+
+        if self.fields_active(self.address_fields):
+            address = obj.addresses[0] if obj.addresses else None
+            for field in self.address_fields:
+                if field in self.fields:
+                    if address:
+                        data[field] = getattr(address, field[8:])
+                    else:
+                        data[field] = None
+
+        return data
+
+    def update_object(self, obj, data, local_data=None):
+        obj = super(ContactMixin, self).update_object(obj, data, local_data)
+
+        if 'phone_number' in self.fields:
+            phones = [phone for phone in obj.phones if phone.type == 'Voice']
+            number = data['phone_number']
+            if number:
+                if phones:
+                    phone = phones[0]
+                    if phone.number != number:
+                        phone.number = number
+                else:
+                    obj.add_phone_number(number, type='Voice')
+            else:
+                for phone in phones:
+                    obj.phones.remove(phone)
+
+        if 'phone_number_2' in self.fields:
+            phones = [phone for phone in obj.phones if phone.type == 'Voice']
+            number = data['phone_number_2']
+            if number:
+                if len(phones) > 1:
+                    phone = phones[1]
+                    if phone.number != number:
+                        phone.number = number
+                else:
+                    obj.add_phone_number(number, 'Voice')
+            else:
+                for phone in phones[1:]:
+                    obj.phones.remove(phone)
+
+        if 'email_address' in self.fields:
+            address = data['email_address']
+            if address:
+                if obj.email:
+                    if obj.email.address != address:
+                        obj.email.address = address
+                else:
+                    obj.add_email_address(address)
+            else:
+                if len(obj.emails):
+                    del obj.emails[:]
+
+        if self.fields_active(self.address_fields):
+            if any([data[f] for f in self.address_fields if f in self.fields]):
+                if obj.addresses:
+                    address = obj.addresses[0]
+                else:
+                    # address = model.CustomerMailingAddress()
+                    address = obj.make_address()
+                    obj.addresses.append(address)
+                for field in self.address_fields:
+                    if field in self.fields:
+                        if getattr(address, field[8:]) != data[field]:
+                            setattr(address, field[8:], data[field])
+            elif obj.addresses:
+                obj.addresses.pop(0)
+
+        return obj
+
+
+class PersonImporter(ContactMixin, ToRattail):
     """
     Person data importer.
     """
     model_class = model.Person
 
+    customer_fields = [
+        'customer_uuid',
+        'customer_person_ordinal',
+    ]
+
+    @property
+    def supported_fields(self):
+        fields = super(PersonImporter, self).supported_fields
+        return fields + self.customer_fields
+
+    def cache_query_options(self):
+        options = super(PersonImporter, self).cache_query_options()
+        if self.fields_active(self.customer_fields):
+            options.append(orm.joinedload(model.Person._customers))
+        return options
+
+    def normalize_local_object(self, person):
+        data = super(PersonImporter, self).normalize_local_object(person)
+
+        if self.fields_active(self.customer_fields):
+            cp = None
+            if person._customers:
+                if len(person._customers) > 1:
+                    raise NotImplementedError("Person has {} customer attachments".format(
+                        len(person._customers)))
+                cp = person._customers[0]
+            if 'customer_uuid' in self.fields:
+                data['customer_uuid'] = cp.customer_uuid if cp else None
+            if 'customer_person_ordinal' in self.fields:
+                data['customer_person_ordinal'] = cp.ordinal if cp else None
+
+        return data
+
+    def update_object(self, person, data, local_data=None):
+        person = super(PersonImporter, self).update_object(person, data, local_data)
+
+        if self.fields_active(self.customer_fields):
+
+            # first we identify the "current" CustomerPerson, if any
+            cp = None
+            if person._customers:
+                if len(person._customers) > 1:
+                    raise NotImplementedError("Person has {} customer attachments".format(
+                        len(person._customers)))
+                cp = person._customers[0]
+
+            if 'customer_uuid' in self.fields:
+                if data['customer_uuid']:
+
+                    # customer identified, so confirm/establish the link
+                    if cp and cp.customer_uuid != data['customer_uuid']:
+                        customer = cp.customer
+                        self.session.delete(cp)
+                        customer._people.reorder()
+                        cp = None
+                    if not cp:
+                        customer = self.session.query(model.Customer)\
+                                               .get(data['customer_uuid'])
+                        cp = model.CustomerPerson()
+                        if not person.uuid:
+                            person.uuid = get_uuid()
+                        cp.person_uuid = person.uuid
+                        customer._people.append(cp)
+
+                    # we only need to process this field if we have a CP record
+                    if cp and 'customer_person_ordinal' in self.fields:
+                        if cp.ordinal != data['customer_person_ordinal']:
+                            customer = cp.customer
+                            cp.ordinal = data['customer_person_ordinal']
+                            customer._people.reorder()
+
+                elif cp:
+                    # host data says no customer, so sever the link
+                    customer = cp.customer
+                    self.session.delete(cp)
+                    customer._people.reorder()
+                    cp = None
+
+        return person
+
 
 class GlobalPersonImporter(GlobalMixin, PersonImporter):
     """
@@ -620,7 +828,7 @@ class WorkedShiftImporter(ToRattail):
     model_class = model.WorkedShift
 
 
-class CustomerImporter(ToRattail):
+class CustomerImporter(ContactMixin, ToRattail):
     """
     Customer data importer.
     """
@@ -630,18 +838,6 @@ class CustomerImporter(ToRattail):
         'first_name',
         'last_name',
     ]
-    phone_fields = [
-        'phone_number',
-        'phone_number_2',
-    ]
-    address_fields = [
-        'address_street',
-        'address_street2',
-        'address_city',
-        'address_state',
-        'address_zipcode',
-        'address_invalid',
-    ]
     group_fields = [
         'group_id',
         'group_id_2',
@@ -649,25 +845,18 @@ class CustomerImporter(ToRattail):
 
     @property
     def supported_fields(self):
-        return self.simple_fields + self.person_fields + self.phone_fields + self.address_fields + self.group_fields + [
-            'email_address',
-        ]
+        fields = super(CustomerImporter, self).supported_fields
+        return fields + self.person_fields + self.group_fields
 
     def setup(self):
         if self.fields_active(self.group_fields):
             self.groups = self.cache_model(model.CustomerGroup, key='id')
 
     def cache_query_options(self):
-        options = []
+        options = super(CustomerImporter, self).cache_query_options()
         if self.fields_active(self.person_fields):
             options.append(orm.joinedload(model.Customer._people)\
                            .joinedload(model.CustomerPerson.person))
-        if self.fields_active(self.phone_fields):
-            options.append(orm.joinedload(model.Customer.phones))
-        if self.fields_active(self.address_fields):
-            options.append(orm.joinedload(model.Customer.addresses))
-        if 'email_address' in self.fields:
-            options.append(orm.joinedload(model.Customer.email))
         if self.fields_active(self.group_fields):
             options.append(orm.joinedload(model.Customer._groups)\
                            .joinedload(model.CustomerGroupAssignment.group))
@@ -684,26 +873,6 @@ class CustomerImporter(ToRattail):
             if 'last_name' in self.fields:
                 data['last_name'] = person.last_name if person else None
 
-        if self.fields_active(self.phone_fields):
-            phones = [phone for phone in customer.phones if phone.type == 'Voice']
-            if 'phone_number' in self.fields:
-                data['phone_number'] = phones[0].number if phones else None
-            if 'phone_number_2' in self.fields:
-                data['phone_number_2'] = phones[1].number if len(phones) > 1 else None
-
-        if 'email_address' in self.fields:
-            email = customer.email
-            data['email_address'] = email.address if email else None
-
-        if self.fields_active(self.address_fields):
-            address = customer.addresses[0] if customer.addresses else None
-            for field in self.address_fields:
-                if field in self.fields:
-                    if address:
-                        data[field] = getattr(address, field[8:])
-                    else:
-                        data[field] = None
-
         if 'group_id' in self.fields:
             group = customer.groups[0] if customer.groups else None
             data['group_id'] = group.id if group else None
@@ -740,60 +909,6 @@ class CustomerImporter(ToRattail):
             if new_person and 'first_name' in self.fields and 'last_name' in self.fields:
                 person.display_name = normalize_full_name(person.first_name, person.last_name)
 
-        if 'phone_number' in self.fields:
-            phones = [phone for phone in customer.phones if phone.type == 'Voice']
-            number = data['phone_number']
-            if number:
-                if phones:
-                    phone = phones[0]
-                    if phone.number != number:
-                        phone.number = number
-                else:
-                    customer.add_phone_number(number, type='Voice')
-            else:
-                for phone in phones:
-                    customer.phones.remove(phone)
-
-        if 'phone_number_2' in self.fields:
-            phones = [phone for phone in customer.phones if phone.type == 'Voice']
-            number = data['phone_number_2']
-            if number:
-                if len(phones) > 1:
-                    phone = phones[1]
-                    if phone.number != number:
-                        phone.number = number
-                else:
-                    customer.add_phone_number(number, 'Voice')
-            else:
-                for phone in phones[1:]:
-                    customer.phones.remove(phone)
-
-        if 'email_address' in self.fields:
-            address = data['email_address']
-            if address:
-                if customer.email:
-                    if customer.email.address != address:
-                        customer.email.address = address
-                else:
-                    customer.add_email_address(address)
-            else:
-                if len(customer.emails):
-                    del customer.emails[:]
-
-        if self.fields_active(self.address_fields):
-            if any([data[f] for f in self.address_fields if f in self.fields]):
-                if customer.addresses:
-                    address = customer.addresses[0]
-                else:
-                    address = model.CustomerMailingAddress()
-                    customer.addresses.append(address)
-                for field in self.address_fields:
-                    if field in self.fields:
-                        if getattr(address, field[8:]) != data[field]:
-                            setattr(address, field[8:], data[field])
-            elif customer.addresses:
-                customer.addresses.pop(0)
-
         if 'group_id' in self.fields:
             group_id = data['group_id']
             if group_id:
@@ -876,7 +991,7 @@ class CustomerPhoneNumberImporter(ToRattail):
     model_class = model.CustomerPhoneNumber
 
 
-class MemberImporter(ToRattail):
+class MemberImporter(ContactMixin, ToRattail):
     """
     Member data importer.
     """
@@ -897,6 +1012,13 @@ class MemberPhoneNumberImporter(ToRattail):
     model_class = model.MemberPhoneNumber
 
 
+class MemberMailingAddressImporter(ToRattail):
+    """
+    Member mailing address data importer.
+    """
+    model_class = model.MemberMailingAddress
+
+
 class VendorImporter(ToRattail):
     """
     Vendor data importer.
@@ -1154,10 +1276,21 @@ class SubdepartmentImporter(ToRattail):
             data['department_number'] = dept.number if dept else None
         return data
 
+    def get_department(self, number):
+        if hasattr(self, 'departments'):
+            return self.departments.get(number)
+
+        try:
+            return self.session.query(model.Department)\
+                               .filter(model.Department.number == number)\
+                               .one()
+        except NoResultFound:
+            pass
+
     def update_object(self, subdepartment, data, local_data=None):
         subdepartment = super(SubdepartmentImporter, self).update_object(subdepartment, data, local_data)
         if 'department_number' in self.fields:
-            dept = self.departments.get(data['department_number'])
+            dept = self.get_department(data['department_number'])
             if not dept:
                 dept = model.Department()
                 dept.number = data['department_number']
