@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 
 from __future__ import unicode_literals, absolute_import
 
@@ -6,8 +6,9 @@ import csv
 import datetime
 import argparse
 from unittest import TestCase
-from cStringIO import StringIO
+from six import StringIO
 
+import six
 # from sqlalchemy import func
 from mock import patch, Mock
 from fixture import TempIO
@@ -51,7 +52,7 @@ class TestCommand(TestCase):
     def test_unicode(self):
         command = core.Command()
         command.name = 'some-app'
-        self.assertEqual(unicode(command), u'some-app')
+        self.assertEqual(six.text_type(command), 'some-app')
         
     def test_iter_subcommands_includes_expected_item(self):
         command = core.Command()
@@ -140,12 +141,6 @@ class TestCommand(TestCase):
 
 class TestSubcommand(TestCase):
 
-    def test_repr(self):
-        command = core.Command()
-        subcommand = core.Subcommand(command)
-        subcommand.name = 'fake-command'
-        self.assertEqual(repr(subcommand), "Subcommand(name=u'fake-command')")
-
     def test_add_parser_args_does_nothing(self):
         command = core.Command()
         subcommand = core.Subcommand(command)
