@@ -1,9 +1,10 @@
-# -*- coding: utf-8 -*-
+# -*- coding: utf-8; -*-
 
 from __future__ import unicode_literals, absolute_import
 
 from unittest import TestCase
 
+import six
 import sqlalchemy as sa
 from fixture import TempIO
 
@@ -77,6 +78,6 @@ class TestConfigExtension(TestCase):
         config.set('rattail.db', 'host.url', host_url)
         db.ConfigExtension().configure(config)
         self.assertEqual(len(config.rattail_engines), 2)
-        self.assertEqual(unicode(config.rattail_engines['default'].url), default_url)
-        self.assertEqual(unicode(config.rattail_engines['host'].url), host_url)
-        self.assertEqual(unicode(config.rattail_engine.url), default_url)
+        self.assertEqual(six.text_type(config.rattail_engines['default'].url), default_url)
+        self.assertEqual(six.text_type(config.rattail_engines['host'].url), host_url)
+        self.assertEqual(six.text_type(config.rattail_engine.url), default_url)
